#pragma once
#include <map>
#include <vector>
#include "out.h"
#include "IRCNsm.h"
#include "IRCCmd.h"
#include "setting_struct.h"

//************************************************

#define DEFINE_FUNC_SWITCH_ELEM(name) \
	void Do##name(LPCSTR from, LPCSTR host, ParamList &param); \
	static void Process##name(CIRCConnection *pThis, LPCSTR from, LPCSTR host, ParamList &param) \
	{ pThis->Do##name(from, host, param); }

//************************************************

class CIRCConnection :
	public COut
{
private:
	typedef std::map<std::string, CIRCSession*> SesMap;
	typedef void (*FuncSwitchFunc)(CIRCConnection*, LPCSTR, LPCSTR, ParamList&);
	struct FuncSwitchElem{
		LPCSTR cmd;
		FuncSwitchFunc func;
		bool operator <(LPCSTR cmp) const
		{ return strcmp(cmd, cmp) < 0; }
		bool operator <(const FuncSwitchElem& fse) const
		{ return strcmp(cmd, fse.cmd) < 0; }
	};

	friend bool operator <(LPCSTR cmp, const FuncSwitchElem& fse);

private:
	CIRCSocket *m_pSocket;
	CIRCCmd m_irc;
	CIRCNsm m_nsm;

	SesMap m_slist;
	ConChList m_chs;
	std::string m_me;
	const std::string m_acc;
	bool m_bConnected;
	HANDLE m_lock;

public:
	CIRCConnection(PLoginInfo pli, HNsmConnection &hCon);
	~CIRCConnection(void);
	CIRCConnection& operator= (CIRCConnection &/*other*/)
	{ _ASSERT(FALSE); return *this; }

	void ChangeUserName(LPCWSTR nick);
	void RemMember(LPCSTR acc);
	void AddGroup(LPCWSTR ch);
	void RemGroup(int nGroupID);

	HNsmSession SesOpen(LPCSTR acc);
	void SesClose(HNsmSession ses);
	void SesInvite(HNsmSession ses, LPCSTR acc);
	void SesSendMsg(HNsmSession ses, LPCWSTR msg);

	void DoExec(LPCSTR line);
	void DoLog(LPCSTR header, LPCSTR body) const;
	void DoError(LPCSTR msg);

	void Connected();
	void Disconnected();
	void ChangeUserName(LPCSTR nick);
	CIRCSession* GetSession(HNsmSession hSession);
	CIRCSession* GetSession(LPCSTR name);
	CIRCSession* CreateSession(LPCSTR ch);
	bool DeleteSession(LPCSTR ch);
	void SessionConnected(LPCSTR ch);
	void SessionMemberAdd(LPCSTR ch, LPCSTR nick, LPCSTR acc);
	void SessionMemberRemove(LPCSTR ch, LPCSTR member);
	void SessionMemberName(LPCSTR oldName, LPCSTR newName);
	void SessionMemberMode(LPCSTR ch, LPCSTR nick, u_int mode);
	void SessionTopic(LPCSTR ch, LPCSTR topic);
	void RecvMessage(LPCSTR from, LPCSTR host, LPCSTR to, LPCSTR msg);

	void Console(LPCSTR line);

	bool IsMe(LPCSTR nick)
	{ return m_me == nick; }

	bool IsChannel(LPCSTR name)
	{ return ((0 < *name) && (*name < 0x41)) || (0x7A < *name); }

private:
	void Login(AccSetting &set, LPCSTR pass);
	void SaveConChList() const;
	void SendMsg(const CIRCSession *s, std::string &str);
	void Lock()
	{ while( ::WaitForSingleObjectEx(m_lock, INFINITE, TRUE) != WAIT_OBJECT_0 ); }
	void Unlock()
	{ ::ReleaseSemaphore(m_lock, 1, NULL); }

private:
	DEFINE_FUNC_SWITCH_ELEM(001) // welcome
	DEFINE_FUNC_SWITCH_ELEM(301) // away
	DEFINE_FUNC_SWITCH_ELEM(305) // unaway
	DEFINE_FUNC_SWITCH_ELEM(306) // now away
	DEFINE_FUNC_SWITCH_ELEM(315) // end of who
	DEFINE_FUNC_SWITCH_ELEM(324) // channel mode is
	DEFINE_FUNC_SWITCH_ELEM(331) // no topic
	DEFINE_FUNC_SWITCH_ELEM(332) // topic
	DEFINE_FUNC_SWITCH_ELEM(341) // inviting
	DEFINE_FUNC_SWITCH_ELEM(352) // who
//	DEFINE_FUNC_SWITCH_ELEM(353) // names
//	DEFINE_FUNC_SWITCH_ELEM(366) // end of names
	DEFINE_FUNC_SWITCH_ELEM(INVITE);
	DEFINE_FUNC_SWITCH_ELEM(JOIN);
	DEFINE_FUNC_SWITCH_ELEM(KICK);
	DEFINE_FUNC_SWITCH_ELEM(KILL);
	DEFINE_FUNC_SWITCH_ELEM(MODE);
	DEFINE_FUNC_SWITCH_ELEM(NICK);
	DEFINE_FUNC_SWITCH_ELEM(NOTICE);
	DEFINE_FUNC_SWITCH_ELEM(PART);
	DEFINE_FUNC_SWITCH_ELEM(PING);
	DEFINE_FUNC_SWITCH_ELEM(PRIVMSG);
	DEFINE_FUNC_SWITCH_ELEM(QUIT);
	DEFINE_FUNC_SWITCH_ELEM(TOPIC);

private:
	static FuncSwitchFunc FSwitch(LPCSTR cmd);
};

inline bool operator <(LPCSTR cmp, const CIRCConnection::FuncSwitchElem& fse)
{
	return strcmp(cmp, fse.cmd) < 0;
}
